
/*
Home Heat recovery Ventilator Controller featuring
Web Server GUI and  NTP service for synchronizing  the ESP32 software RTC
web server shows the temperature/humidity of one CC2A23 sensor (indoors) and a thermistor (outdoors)
as well as the local time
and the 24 hour slots for heat recovery ventilator control
using a ESP32 dev board from Adafruit
CC2A23 temp/humidity Sensor (amphenol advanced sensors) to pins IO32, 33   (ADC1 chan 4,5,)
outdoor temperature sensed by 10K thermistor/10K resistor to 3.3V  on IO34 (ADC1 chan 6)
Relay output on IO2
Sound output on I21
// August 22 Added a check for network connection every minute- and a re connect if necessary
// Added a NTP request every day at 2 PM to sync the clock
// Time updated via serial.println once per minute
*/

#include <EEPROM.h>
#include <WiFi.h>
#include <ESPmDNS.h>
#include <WiFiUdp.h>
#include <WiFiClient.h>
#include <sys/time.h>
#include <esp32-hal-adc.h>

#define EEPROM_SIZE 100\

char ssid[] = "xxxxxxxx";  //  your network SSID (name)
char pass[] = "xxxxxxxx";       // your network password

#define DNSname "esp32"  

byte mac[6];                     // the MAC address of your Wifi shield
int EEPROM_addr=0;
IPAddress timeServer(132, 163, 4, 101); // time-a.timefreq.bldrdoc.gov NTP server

unsigned int localPort = 8888;      // local port to listen for UDP packets
const int NTP_PACKET_SIZE= 48; // NTP time stamp is in the first 48 bytes of the message
byte packetBuffer[ NTP_PACKET_SIZE]; //buffer to hold incoming and outgoing packets 
unsigned long epoch2;
byte currentHour;
byte currentMinute;
byte hour;
byte minute;
byte second;
int tick;
byte action;
String readString;
String targetStr;
byte season;
int RHsetpoint = 50;
float insideTemperature;
float outsideTemperature;
float insideHumidity;
float outsideHumidity;
byte overRide = 0;
int relay =2;  // output relay connected to DIGITAL 2
int sound = 21;  // piezo on I21
String timeASCII;
uint16_t ADC;

struct timeval currentTime;

// instantiate a UDP object to let us send and receive NTP packets over UDP
WiFiUDP Udp;

// Initialize the Ethernet server library
// with the IP address and port 80
WiFiServer server(80);
WiFiClient client = server.available();

void setup() {
readString.reserve(250);  // we won't read in any more than 200 chars from input stream
pinMode(relay,OUTPUT);
pinMode(sound, OUTPUT);

digitalWrite(relay,HIGH);       // toggle relay for 2 secs for testing  
delay(2000);
digitalWrite(relay,LOW);      // relay is active high- de-activate it
beep();  // sound piezo alarm for a few secs to test it

 // Open serial communications 
  Serial.begin(115200);
  Serial.println("Heat Recovery ventilator Controller");
  analogSetAttenuation(ADC_11db);  // full scale = 3.64 volts (but limited to 3.3 V of ESP32 VDD supply)
  analogSetCycles(32);  // average 32 samples per call 
    
  Serial.print("Connecting to ");
  Serial.println(ssid);

   // start wifi and UDP
    WiFi.begin(ssid, pass);
    while (WiFi.status() != WL_CONNECTED) {
      delay(500);
      Serial.print(".");
    }
    Serial.println("");
    Serial.println("WiFi connected");
    // Print the IP address
    Serial.print("IP# = ");
    Serial.println(WiFi.localIP());
 
   if (!MDNS.begin(DNSname)) {                            
		Serial.println("Error setting up MDNS responder!");
		while (1) {
			delay(1000);
			beep();
		}
	}
	Serial.println("mDNS responder started");


	GetNTPTime(); // get time from network via SNTP protocol
//	analogSetAttenuation(ADC_11db);  // full scale = 3.6V power upply
//	analogSetCycles(32);  // get 32 samples averaged
   if(!EEPROM.begin(EEPROM_SIZE)) {
    while(1) 
     Serial.println("EEPROM failed to initialize");
	beep();
	delay(1000);
	beep();
	delay(6000);
  }

  // start the server:

  server.begin();
  Serial.print("webserver is at ");
  Serial.println(WiFi.localIP());
  season = EEPROM.read(25);
  RHsetpoint = EEPROM.read(26);
  tick = 0;
}

void loop() {
	tick++;
	if (tick > 60000) {    // about once per minute
		convertTime(); // get time from ESP32 internal RTC and set vars hour,minute,second
		tick = 0;
		if (WiFi.status() != WL_CONNECTED) {
			while (WiFi.status() != WL_CONNECTED) {
				WiFi.begin(ssid, pass);
				delay(500);
				beep();
				Serial.print(".");
			}
			Serial.println("Reconnected");
			if (!MDNS.begin("esp32")) {                             // used on line 404 as well !!!
				Serial.println("Error setting up MDNS responder!");
			}
			server.begin();
			Serial.print("webserver is at ");
		}
	}
   delay(1); 
   if (currentHour != hour){  
    currentHour = hour;
	if (currentHour == 14) {
		GetNTPTime();  // non blocking if no NTP response
	}
    if (overRide == 0) {    // at the end of the override interval, turn off the H.E.
    digitalWrite(relay,LOW);
    }      
    Serial.print("hour boundary=  ");
    Serial.println(currentHour );
    action=EEPROM.read(currentHour);
    if (action == 255) {
      Serial.println (" OFF");
      digitalWrite(relay,LOW);
    }
    if (action ==1) {
     Serial.println(" ON"); 
     digitalWrite(relay,HIGH);
    }
    if (action==2){
    Serial.println(" Favored");
     if ( season ==255 ){
      //season is winter
      if ( (outsideTemperature > 10) && (insideHumidity > RHsetpoint)) {  
          digitalWrite(relay,HIGH);
          Serial.println(" ON"); 
      }
     } 
     if (season ==1) {
      // season is summer
      if ((outsideTemperature <28) && (insideHumidity > RHsetpoint) ) {
          digitalWrite(relay,HIGH);
          Serial.println(" ON"); 
      }
     }  
   }
   if (overRide >0) {
    overRide--;
   } 
  }
  if (overRide > 0) {
    digitalWrite(relay,HIGH);
    Serial.println("overRide");
  }  

  // listen for incoming clients

  client = server.available();

  if (client) {
     readString = "";    
     Serial.println("new client");
    // an http request ends with a blank line
    boolean currentLineIsBlank = true;
    while (client.connected()) {
      if (client.available()) {
        char c = client.read();
        if (readString.length() <200) {
          readString += c;
        }
        if (c== '\n') {
           //Serial.print(readString);
          int indx0 =readString.indexOf("GET");
          int indx1 =readString.indexOf ("?ON+time=");
          int indx2 =readString.indexOf("&",indx1);
          if (indx0 ==0  && indx2 > indx1+9) {
           String tmp1 = readString.substring(indx1+9,indx2); 
           int onTime= tmp1.toInt();
           if (onTime >= 0 && onTime <24) {
           EEPROM.write(onTime,1); 
		   EEPROM.commit();
           }  
          }         
          int indx3 =readString.indexOf ("OFF+time=");
          int indx4 =readString.indexOf("&",indx3);
          if (indx0 ==0 &&  indx4 > indx3+9) {
           String tmp2 = readString.substring(indx3+9,indx4);   
           int offTime= tmp2.toInt();
           if (offTime >= 0 && offTime <24) {
           EEPROM.write(offTime,255); 
		   EEPROM.commit();
           }  
          }
          int indx5 =readString.indexOf ("FAVORED+time=",indx4);
          int indx6 =readString.indexOf("&",indx5);
          if (indx0 ==0 && indx6 > indx5+13) {
           String tmp3= readString.substring(indx5+13,indx6);
           int favoredTime= tmp3.toInt();
           if (favoredTime >= 0 && favoredTime <24) {
           EEPROM.write(favoredTime,2);
		   EEPROM.commit();
           }  
          }
           int indx7 =readString.indexOf ("Clear+All+Time+Slots",0);
          if (indx0 ==0 && indx7 >1) {
            // clear all time slots to "off" in EEPROM
            for ( int j = 0; j < 24; j++){
              EEPROM.write(j,255);
			  EEPROM.commit();
            }    
          }  
          int indx8 =readString.indexOf ("group1=Winter",0);          
         if (indx0 ==0 && indx8>1) {
          season=255; 
          EEPROM.write(25,255);
		  EEPROM.commit();
         }
          int indx9 =readString.indexOf ("group1=Summer",0);          
         if (indx0 ==0 && indx9>1) {
          season=1; 
          EEPROM.write(25,1);
		  EEPROM.commit();
         }
         
          int indx10 =readString.indexOf ("RHSetpoint=");
          int indx11 =readString.indexOf ("&",indx10);
          if (indx0 ==0 && (indx11 > indx10 +12) ) {
           String tmp3= readString.substring(indx10+11,indx11);
           int RHtemp= tmp3.toInt();
           if ((RHtemp >= 30) && (RHtemp <70) ) {
            RHsetpoint= RHtemp;
            EEPROM.write(26,RHsetpoint); 
			EEPROM.commit();
           }  
         }

           int indx12 =readString.indexOf ("OverRide",0);
          if (indx0 ==0 && indx12 >1) {
            // set the OverRide value to 3 (turns H.E. on for 3 hours regardless of settings
           overRide=3;
           Serial.println("overRide on");
          }           
          
         readString=""; 
        } 
        // if you've gotten to the end of the line (received a newline
        // character) and the line is blank, the http request has ended,
        // so you can send a reply
        if (c == '\n' && currentLineIsBlank) {
          // send a standard http response header
          client.println("HTTP/1.1 200 OK");
          client.println("Content-Type: text/html");
          client.println("Connection: close");  // the connection will be closed after completion of the response
//	  client.println("Refresh: 20");  // refresh the page automatically every 10 sec
          client.println();
          client.println("<!DOCTYPE HTML>");
          client.println("<html>");
          client.println("<br />"); 
          client.print("Time = ");       // UTC is the time at Greenwich Meridian (GMT)
          client.print(hour); 
          client.print(":");
          if (minute < 10)
           client.print('0');
          client.print(minute);
          client.print(":");
          if (second < 10)
           client.print('0');
          client.print(second);
          client.print("&nbsp");
          client.print("&nbsp");
          client.print("&nbsp");
          client.print("Season = ");
          if (season == 255) {
            client.println("Winter");
          }
          else {
           client.println("Summer");
          } 
          client.print("&nbsp");
          client.print("&nbsp");
          client.print("RH setpoint=");
          client.print(RHsetpoint);
          client.print("&nbsp");
          client.print("&nbsp");
          client.print("overRide Hours =");
          client.println(overRide);
          client.println("<br />"); 
          client.println("Time slots");
          client.println("<table border= '1'");
          client.println("cellpadding= '10'>");
          client.println("<tr>");
          for (int hr =0; hr <12;hr++) {
          client.print("<td>");
          client.print(hr);
          client.println("</td>");
          }
          client.println("</tr>");
          for (int hr =0; hr <12;hr++) {
           client.println("<td>");
           if (EEPROM.read(hr) == 255) {
             client.print("OFF");
             client.print("</td>");
           }
           if (EEPROM.read(hr)== 1) {
             client.print("ON");
             client.print("</td>");
           }
           if (EEPROM.read(hr) == 2) {
             client.print("FAV");
             client.print("</td>");
           }
          }                
          client.println("</tr>");
          client.println("</table>");
          client.println("<br />");
          client.println("<table border='1'");
          client.println("cellpadding= '10'>");
          client.println("<tr>");
          for (int hr =12; hr <24;hr++) {
          client.print("<td>");
          client.print(hr);
          client.println("</td>");
          }
          client.println("</tr>");
          for (int hr =12; hr <24;hr++) {
           client.println("<td>");
           if (EEPROM.read(hr) == 255) {
             client.print("OFF");
             client.print("</td>");
           }
           if (EEPROM.read(hr) == 1) {
             client.print("ON");
             client.print("</td>");
           }
           if (EEPROM.read(hr) == 2) {
             client.print("FAV");
             client.print("</td>");
           }
          }
          client.println("</tr>");
          client.println("</table>");
          client.println("<br />"); 

// get the inside temperature and humidity and outside temperature from sensors
		  ADC = analogRead(32);
		  //insideHumidity = ((ADC/4096.0)*3510) / 33.0;   // FS mv of ADC = 3.51V or 3510 mv sensor VDD=3.3V
		  insideHumidity = (ADC/1.124) / 33.0;   // FS mv of ADC = 3.51V or 3510 mv sensor VDD=3.3V
		  Serial.print("Inside humidity");
		  Serial.println(insideHumidity);
		  ADC = analogRead(33);
		  // sensor temperature = (reading[mv] /20)  -40.0  from Chipcap  datasheet page 27
		  insideTemperature = ((ADC/1.124)/20) - 40.0;   // 1.124 is a constant derived from measured values with atten= 11 dB
		  //Serial.print("ADC=");
		  //Serial.println(ADC);
		  Serial.print("inside Temperature");
		  Serial.println(insideTemperature);
		  ADC = analogRead(34);
		  float resistance = (10000.0*ADC / 1125.0) / (3.3 - ADC / 1125.0);    // 1125.0 = 4096/3.64 Vfs
		  // Steinhart-hart equation
		  // B for BC2301 (NTCLE100E3103JB0) thermistor is 3977
		  float temperature = (1 / 298.15) + log(resistance / 10000.0) / 3977.0;
		  outsideTemperature = (1 / temperature) - 273;
		  Serial.print("outside temperature ");
		  Serial.println(outsideTemperature);
//		  delay(1000);
            client.print("Inside Temperature = ");
            client.print(insideTemperature);
			client.println("<br />");
			client.print("Inside Humidity = ");
			client.print(insideHumidity);
			client.println("<br />");
			client.print("Outside Temperature = ");
			client.print(outsideTemperature);
            client.println("<br />");
            client.println("<br />");
          //  client.println("<FORM ACTION=\"http://esp32.local:80\" method=get >");
            client.print("FORM ACTION = \"http://");
            client.print(DNSname);
            client.println(".local:80\" method=get >");
            client.println("Time in 24H clock: <INPUT TYPE=TEXT NAME='ON time' VALUE='' SIZE='6' MAXLENGTH='5'>");
            client.println("<INPUT TYPE=SUBMIT NAME='submit' VALUE= 'ON time'> <BR>");
            client.println("Time in 24H clock: <INPUT TYPE=TEXT NAME='OFF time' VALUE='' SIZE='6' MAXLENGTH='5'>");
            client.println("<INPUT TYPE=SUBMIT NAME='submit' VALUE= 'OFF time'><BR>");
            client.println("Time in 24H clock: <INPUT TYPE=TEXT NAME='FAVORED time' VALUE='' SIZE='6' MAXLENGTH='5'>");
            client.println("<INPUT TYPE=SUBMIT NAME='submit' VALUE= 'FAVORED time'><BR>"); 
            client.println("RH setpoint: <INPUT TYPE=TEXT NAME='RHSetpoint' VALUE='' SIZE='6' MAXLENGTH='5'>");
            client.println("<INPUT TYPE=SUBMIT NAME='submit' VALUE= 'RHSetpoint'><BR>"); 
            client.println("<INPUT TYPE=SUBMIT NAME='submit' VALUE= 'Clear All Time Slots'>");
            client.print("&nbsp");
            client.print("&nbsp");
            client.print("&nbsp");            
            client.println("<input type= 'radio' name='group1' value='Summer'> Summer");
            client.println("<input type= 'radio' name='group1' value='Winter' > Winter  ");
            client.println("<INPUT TYPE=SUBMIT NAME='submit' VALUE= 'Set Season'>");
            client.print("&nbsp");
            client.print("&nbsp");
            client.print("&nbsp");
            client.println("<INPUT TYPE=SUBMIT NAME='submit' VALUE= 'OverRide'>");
            client.println("</html>");
          break;
        }
        if (c == '\n') {
          // you're starting a new line
          currentLineIsBlank = true;
        } 
        else if (c != '\r') {
          // you've gotten a character on the current line
          currentLineIsBlank = false;
        }
      }
    }
    // give the web browser time to receive the data
    delay(1);
    // close the connection:
    client.stop();
    Serial.println("client disconnected");
  }
}


void GetNTPTime()  
{
 //  Udp.begin(localPort);
  sendNTPpacket(timeServer); // send an NTP packet to a time server
    // wait to see if a reply is available
  delay(5000);  
  if ( Udp.parsePacket() ) { 
	  Serial.println("NTP packet received");
    // We've received a packet, read the data from it
    Udp.read(packetBuffer,NTP_PACKET_SIZE);  // read the packet into the buffer
    //the timestamp starts at byte 40 of the received packet and is four bytes,
    // or two words, long. First, esxtract the two words:
    unsigned long highWord = word(packetBuffer[40], packetBuffer[41]);
    unsigned long lowWord = word(packetBuffer[42], packetBuffer[43]);  
    // combine the four bytes (two words) into a long integer
    // this is NTP time (seconds since Jan 1 1900):
    unsigned long secsSince1900 = highWord << 16 | lowWord;  
    // now convert NTP time into everyday time:
    // Unix time starts on Jan 1 1970. In seconds, that's 2208988800:
    const unsigned long seventyYears = 2208988800UL;     
    // subtract seventy years:
    unsigned long epoch = secsSince1900 - seventyYears; 
   epoch2=epoch - 10800;  // epoch2 compensates for timezone & DST in Atlantic Canada 3 hr
   unsigned long x = (epoch2 % 86400L); // convert to elapsed seconds today
   currentTime.tv_sec = x;    // set the ESP32 time clock to this walue
   int status = settimeofday(&currentTime, NULL);
   convertTime();
   //delay(2000);
	}
  }


// send an NTP request to the time server at the given address 
unsigned long sendNTPpacket(IPAddress& address)
{
  // set all bytes in the buffer to 0
  memset(packetBuffer, 0, NTP_PACKET_SIZE); 
  // Initialize values needed to form NTP request
  // (see URL above for details on the packets)
  packetBuffer[0] = 0b11100011;   // LI, Version, Mode
  packetBuffer[1] = 0;     // Stratum, or type of clock
  packetBuffer[2] = 6;     // Polling Interval
  packetBuffer[3] = 0xEC;  // Peer Clock Precision
  // 8 bytes of zero for Root Delay & Root Dispersion
  packetBuffer[12]  = 49; 
  packetBuffer[13]  = 0x4E;
  packetBuffer[14]  = 49;
  packetBuffer[15]  = 52;

  // all NTP fields have been given values, now
  // you can send a packet requesting a timestamp: 		   
  Udp.beginPacket(address, 123); //NTP requests are to port 123
  Udp.write(packetBuffer,NTP_PACKET_SIZE);
  Udp.endPacket(); 
  Serial.println("NTP packet sent");
}

void convertTime() {
	int status = gettimeofday(&currentTime, NULL);
	unsigned long x = (currentTime.tv_sec % 86400) ;  //86400 is secs/day
		// print the hour, minute and second:
	Serial.print("Local time is ");       // UTC is the time at Greenwich Meridian (GMT)
	hour = (x / 3600);
	Serial.print(hour); // 
	Serial.print(':');
	minute = (x % 3600)/60;
	if (( minute) < 10) {
		// In the first 10 minutes of each hour, we'll want a leading '0'
		Serial.print('0');
	}
	Serial.print(minute); // print the minute (3600 equals secs per minute)
	Serial.print(':');
	second = (x % 60);
	if (second < 10) {
		// In the first 10 seconds of each minute, we'll want a leading '0'
		Serial.print('0');
	}
	Serial.println(second); // print the second
}

void beep() {
	for (int i = 0; i < 1000; i++) {
		digitalWrite(sound, LOW);
		delay(1);
		digitalWrite(sound, HIGH);
		delay(1);
	}

}


